SystemOrganization addCategory: #'Neo-Caching'!

TestCase subclass: #NeoLRUCacheTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Caching'!

!NeoLRUCacheTests commentStamp: '<historical>' prior: 0!
I am NeoLRUCacheTests.
!

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/28/2013 23:07'!
test5kClasses
	| cache data |
	(cache := NeoLRUCache new)
		maximumWeight: 1024.
	data := Object allSubclasses first: 5*1024.
	data do: [ :each |
		cache at: each name ifAbsentPut: [ each ] ].
	cache validateInvariantWith: self.
	self assert: cache size equals: 1024.
	data shuffled do: [ :each |
		cache at: each name ifAbsentPut: [ each ] ].
	self assert: cache size equals: 1024.
	cache validateInvariantWith: self.	
	data 
		select: [ :each | cache includesKey: each name ]
		thenDo: [ :each | cache at: each name ifAbsentPut: [ self fail ] ].
	self assert: cache hits >= 1024.
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/28/2013 22:52'!
test6k
	| cache |
	(cache := NeoLRUCache new)
		maximumWeight: 600.
	1 to: 6000 do: [ :each |
		cache at: each ifAbsentPut: [ each * 2 ] ].
	self assert: cache size equals: 600.
	self assert: (cache includesKey: 5401).
	self deny: (cache includesKey: 5400).
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/28/2013 16:03'!
testCustomWeight
	| cache |
	(cache := NeoLRUCache new)
		computeWeight: #sizeInMemory;
		maximumWeight: 64.
	cache at: 1 ifAbsentPut: [ ByteArray new: 1 ].
	self assert: cache totalWeight equals: (ByteArray new: 1) sizeInMemory.
	2 to: 10 do: [ :each | 
		cache at: each ifAbsentPut: [ ByteArray new: each ] ].
	self assert: cache size equals: 3.
	self 
		assert: cache totalWeight 
		equals: ((8 to: 10) collect: [ :each | (ByteArray new: each) sizeInMemory ]) sum.
	self assert: (cache at: 10 ifAbsentPut: [ self fail ]) equals: (ByteArray new: 10).
	self deny: (cache includesKey: 1)! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/27/2013 10:14'!
testEmpty
	| emptyCache |
	emptyCache := NeoLRUCache new.
	self assert: emptyCache isEmpty.
	self assert: emptyCache size isZero.
	self assert: emptyCache hits isZero.
	self assert: emptyCache misses isZero.
	self assert: emptyCache hitRatio isZero.
	self assert: emptyCache totalWeight isZero.
	self deny: (emptyCache includesKey: #foo).
	emptyCache keysAndValuesDo: [ :key :value | self fail ]! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/29/2013 12:01'!
testEnumeration
	| cache data keys values |
	cache := NeoLRUCache new.
	data := Dictionary new.
	1 to: 10 do: [ :each | data at: each asWords put: each ].
	data keysAndValuesDo: [ :key :value |
		cache at: key ifAbsentPut: [ value ] ].
	keys := Array new writeStream.
	values := Array new writeStream.
	cache keysAndValuesDo: [ :key :value |
		self assert: (data at: key) equals: value.
		keys nextPut: key.
		values nextPut: value ].
	self assert: keys contents asSet equals: data keys asSet.
	self assert: values contents asSet equals: data values asSet.
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/28/2013 22:51'!
testEviction
	| cache |
	cache := NeoLRUCache new.
	cache maximumWeight: 16.
	1 to: 20 do: [ :each |
		cache at: each asWords ifAbsentPut: [ each ] ].
	self assert: cache size equals: 16.
	self assert: cache totalWeight equals: 16.
	5 to: 20 do: [ :each |
		self assert: (cache includesKey: each asWords).
		self 
			assert: (cache at: each asWords ifAbsentPut: [ self fail ]) 
			equals: each ].
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/29/2013 21:05'!
testFactory
	| cache |
	cache := NeoLRUCache new.
	cache maximumWeight: 5.
	cache factory: [ :key | key * 2 ].
	#( 1 2 3 4 1 5 6 7 8 1 ) do: [ :each |
		 cache at: each ].
	self assert: cache size equals: 5.
	self assert: cache hits equals: 2.
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/29/2013 21:04'!
testFactoryStyle
	| cache factory |
	cache := NeoLRUCache new.
	cache maximumWeight: 5.
	factory := [ :key | key * 2 ].
	#( 1 2 3 4 1 5 6 7 8 1 ) do: [ :each |
		 cache at: each ifAbsentPut: factory ].
	self assert: cache size equals: 5.
	self assert: cache hits equals: 2.
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/29/2013 11:08'!
testFixedAccess
	| cache data keys |
	(cache := NeoLRUCache new)
		maximumWeight: 4.
	data := (1 to: 16) collect: #asWords.
	data do: [ :each |
		cache at: each ifAbsentPut: [ each ] ].
	cache validateInvariantWith: self.
	self assert: cache size equals: 4.
	(#(14 15) collect: #asWords) do: [ :each |
		cache at: each ifAbsentPut: [ each ] ].
	self assert: cache size equals: 4.
	cache validateInvariantWith: self.	
	keys := data select: [ :each | cache includesKey: each ].
	keys do: [ :each | cache at: each ifAbsentPut: [ self fail ] ].
	self assert: cache hits >= 4.
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/28/2013 22:48'!
testOne
	| cache |
	cache := NeoLRUCache new.
	cache at: #foo ifAbsentPut: [ 100 ].
	self assert: (cache includesKey: #foo).
	self deny: cache isEmpty.
	self assert: cache size equals: 1.
	self assert: cache totalWeight equals: 1.
	self assert: cache hits isZero.
	self assert: cache misses equals: 1.
	self assert: cache hitRatio isZero.
	self deny: (cache includesKey: #bar).
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/28/2013 22:50'!
testOneHit
	| cache value |
	cache := NeoLRUCache new.
	cache at: #foo ifAbsentPut: [ 100 ].
	self assert: (cache includesKey: #foo).
	self assert: cache hits equals: 0.
	self assert: cache misses equals: 1.
	value := cache at: #foo ifAbsentPut: [ self fail ].
	self assert: value equals: 100.
	self assert: cache hits equals: 1.
	self assert: cache misses equals: 1.
	cache validateInvariantWith: self
	! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/28/2013 22:50'!
testOneHitTwice
	| cache value |
	cache := NeoLRUCache new.
	cache at: #foo ifAbsentPut: [ 100 ].
	value := cache at: #foo ifAbsentPut: [ self fail ].
	self assert: value equals: 100.
	value := cache at: #foo ifAbsentPut: [ self fail ].
	self assert: value equals: 100.
	self assert: cache hitRatio equals: 2/3.
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/1/2013 16:51'!
testPrimeFactors
	"[ self run: #testPrimeFactors ] bench."
	
	| cache data |
	cache := NeoLRUCache new.
	cache maximumWeight: 512.
	cache factory: [ :key | key ].
	data := Array streamContents: [ :out |
		1 to: 4096 do: [ :each |
			each primeFactorsOn: out.
			out nextPut: each ] ].
	data := data collect: [ :each | each asWords ].
	data do: [ :each | cache at: each ].
	self assert: cache totalWeight equals: 512.
	self assert: cache hitRatio > (7/10).
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/28/2013 23:28'!
testRandomAccess
	| cache data random keys |
	(cache := NeoLRUCache new)
		maximumWeight: 4.
	data := (1 to: 16) collect: #asWords.
	data do: [ :each |
		cache at: each ifAbsentPut: [ each ] ].
	cache validateInvariantWith: self.
	self assert: cache size equals: 4.
	random := data shuffled.
	random do: [ :each |
		cache at: each ifAbsentPut: [ each ] ].
	self assert: cache size equals: 4.
	cache validateInvariantWith: self.	
	keys := data select: [ :each | cache includesKey: each ].
	keys do: [ :each | cache at: each ifAbsentPut: [ self fail ] ].
	self assert: cache hits >= 4.
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/28/2013 22:50'!
testRemoveAll
	| cache |
	cache := NeoLRUCache new.
	1 to: 10 do: [ :each | 
		cache at: each ifAbsentPut: [ each ] ].
	cache removeAll.
	self assert: cache isEmpty.
	self assert: cache size isZero.
	self assert: cache totalWeight isZero.
	self assert: cache misses equals: 10.
	self assert: cache hits isZero.
	self deny: (cache includesKey: #1).
	cache keysAndValuesDo: [ :key :value | self fail ].
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/28/2013 22:50'!
testRemoveOne
	| cache |
	cache := NeoLRUCache new.
	cache at: #foo ifAbsentPut: [ 100 ].
	cache removeKey: #foo.
	self assert: cache isEmpty.
	self assert: cache size isZero.
	self assert: cache totalWeight isZero.
	self deny: (cache includesKey: #foo).
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/28/2013 22:51'!
testRemoveOneOfThree
	| cache |
	cache := NeoLRUCache new.
	cache at: #x ifAbsentPut: [ 100 ].
	cache at: #y ifAbsentPut: [ 200 ].
	cache at: #z ifAbsentPut: [ 300 ].
	cache removeKey: #y.
	self assert: cache size equals: 2.
	self assert: cache totalWeight equals: 2.
	self assert: (cache at: #x ifAbsentPut: [ self fail ]) equals: 100.
	self deny: (cache includesKey: #y).
	self assert: (cache at: #z ifAbsentPut: [ self fail ]) equals: 300.
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/28/2013 22:51'!
testTen
	| cache |
	cache := NeoLRUCache new.
	1 to: 10 do: [ :each |
		cache at: each asWords ifAbsentPut: [ each ] ].
	self assert: cache size equals: 10.
	1 to: 10 do: [ :each |
		self assert: (cache includesKey: each asWords) ].
	1 to: 10 do: [ :each |
		self 
			assert: (cache at: each asWords ifAbsentPut: [ self fail ]) 
			equals: each ].
	self assert: cache hitRatio equals: 1/2.
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/28/2013 22:51'!
testThree
	| cache |
	cache := NeoLRUCache new.
	cache at: #x ifAbsentPut: [ 100 ].
	cache at: #y ifAbsentPut: [ 200 ].
	cache at: #z ifAbsentPut: [ 300 ].
	self assert: (cache at: #x ifAbsentPut: [ self fail ]) equals: 100.
	self assert: (cache at: #y ifAbsentPut: [ self fail ]) equals: 200.
	self assert: (cache at: #z ifAbsentPut: [ self fail ]) equals: 300.
	self assert: cache hitRatio equals: 3/6.
	self assert: (cache at: #z ifAbsentPut: [ self fail ]) equals: 300.
	self assert: (cache at: #y ifAbsentPut: [ self fail ]) equals: 200.
	self assert: (cache at: #x ifAbsentPut: [ self fail ]) equals: 100.
	self assert: cache hitRatio equals: 6/9.
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/28/2013 22:51'!
testThreeHitSameOne
	| cache |
	cache := NeoLRUCache new.
	cache at: #x ifAbsentPut: [ 100 ].
	cache at: #y ifAbsentPut: [ 200 ].
	cache at: #z ifAbsentPut: [ 300 ].
	self assert: (cache at: #x ifAbsentPut: [ self fail ]) equals: 100.
	self assert: (cache at: #y ifAbsentPut: [ self fail ]) equals: 200.
	self assert: (cache at: #z ifAbsentPut: [ self fail ]) equals: 300.
	3 timesRepeat: [ 	
		self assert: (cache at: #x ifAbsentPut: [ self fail ]) equals: 100 ].
	self assert: cache hits equals: 6.
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/28/2013 22:51'!
testTwo
	| cache |
	cache := NeoLRUCache new.
	cache at: #foo ifAbsentPut: [ 100 ].
	cache at: #bar ifAbsentPut: [ 200 ].
	self assert: (cache at: #foo ifAbsentPut: [ self fail ]) equals: 100.
	self assert: (cache at: #bar ifAbsentPut: [ self fail ]) equals: 200.
	self assert: cache hitRatio equals: 2/4.
	self assert: (cache at: #bar ifAbsentPut: [ self fail ]) equals: 200.
	self assert: (cache at: #foo ifAbsentPut: [ self fail ]) equals: 100.
	self assert: cache hitRatio equals: 4/6.
	cache validateInvariantWith: self! !

Object subclass: #NeoLRUCache
	instanceVariableNames: 'lruHead lruTail keyIndex totalWeight maximumWeight computeWeight hits misses factory'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Caching'!

!NeoLRUCache commentStamp: '<historical>' prior: 0!
I am NeoLRUCache.

I am a limited cache that evicts the least recently used entries. My implementation is properly O(1).

My primary interface is #at:ifAbsentPut: which takes two arguments: a key and a block. Either the key is found (cache hit) and its value is returned, or the key is not found (cache miss). If the latter case, block should compute a new value to cache. Because block takes the key as optional argument, you can specify a factory style argument as well. With an explicit factory specified, you can also use #at: to access me.

For each addition to the cache, a weight is computed by #computeWeight (a selector or block) and added to #totalWeight. When totalWeight is no longer below maximumWeight, the least recently used item of the cache is evicted (removed) to make room. 

The default #computeWeight returns 1 for each value, effectively counting the number of entries. The default #maximumWeight is 16.

I count hits and misses and can return my #hitRatio.

Implementation Notes

The key/value pairs in the cache are held as Associations in a linked list, using ValueLink's, between lruHead and lruTail, ordered from least to most recently used.

The keyIndex Dictionary maps from each key to the predecessor ValueLink of the  ValueLink holding the matching key/value pair. To refer to the first element, lruHead, nil is used as predecessor.

New pairs are added at the end of the list.

In case of a hit, a pair gets promoted to the end of the list.

In case of a full cache, the first pair of the list gets evicted.

When manipulating the linked list, the predecessor pointers in keyIndex must be updated where necessary.

See #validateInvariantWith: where the relationship between the 2 datastructures is checked.!

!NeoLRUCache methodsFor: 'private' stamp: 'SvenVanCaekenberghe 11/30/2013 13:04'!
addWeightOf: value 
	| weight |
	weight := computeWeight cull: value.
	totalWeight := totalWeight + weight.
	[ totalWeight > maximumWeight ]
		whileTrue: [ self evict ]! !

!NeoLRUCache methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/29/2013 21:02'!
at: key
	"If key is present in the cache, return the associated value.
	This is a hit and makes that key/value pair the most recently used.
	If key is absent, use the factory to compute a new value and cache it.
	This is a miss and will create a new key/value pair entry.
	Furthermore this could result in the least recently used key/value pair 
	being removed when the specified maximum cache weight is exceeded.
	A factory is required for this behavior."
	
	^ self 
		at: key 
		ifAbsentPut: (factory ifNil: [ self error: 'No factory specified' ])! !

!NeoLRUCache methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/30/2013 19:21'!
at: key ifAbsentPut: block
	"If key is present in the cache, return the associated value.
	This is a hit and makes that key/value pair the most recently used.
	If key is absent, use block to compute a new value and cache it.
	Block can optionally take one argument, the key. 
	This is a miss and will create a new key/value pair entry.
	Furthermore this could result in the least recently used key/value pair 
	being removed when the specified maximum cache weight is exceeded."

	| association |
	association := keyIndex 
		associationAt: key 
		ifAbsent: [ 
			| value |
			value := block cull: key.
			^ self handleMiss: key -> value ].
	^ self handleHit: association
		! !

!NeoLRUCache methodsFor: 'initialize' stamp: 'SvenVanCaekenberghe 11/29/2013 21:11'!
computeWeight: valuable
	"Set the way to compute the weight of each cached value.
	This can be either a Symbol or one argument block.
	When the total weight is no longer below the maximum weight,
	the least recently used key/value pair will be removed.
	The default way to compute the weight returns 1 for each value,
	effectively counting the number of cached values."
	
	computeWeight := valuable! !

!NeoLRUCache methodsFor: 'private' stamp: 'SvenVanCaekenberghe 11/30/2013 19:27'!
evict
	| link value |
	link := self removeFirstLink.
	value := link value.
	self removeWeightOf: value value.
	keyIndex removeKey: value key.
	lruHead 
		ifNotNil: [ keyIndex at: lruHead value key put: nil ]! !

!NeoLRUCache methodsFor: 'initialize' stamp: 'SvenVanCaekenberghe 11/29/2013 20:55'!
factory: block
	"Set the factory to compute values from keys to block.
	The factory will be evaluated for each key not present.
	Only my #at: message will use the factory."
	
	factory := block! !

!NeoLRUCache methodsFor: 'private' stamp: 'SvenVanCaekenberghe 11/30/2013 19:18'!
handleHit: association
	| predecessor newPredecessor link |
	hits := hits + 1.
	predecessor := association value.
	link := predecessor 
		ifNil: [ lruHead ] 
		ifNotNil: [ predecessor nextLink ].
	newPredecessor := self promoteSuccessorOf: predecessor.
	association value: newPredecessor.
	^ link value value! !

!NeoLRUCache methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/1/2013 00:24'!
handleMiss: association
	| link predecessor |
	misses := misses + 1.
	link := association asLink.
	self addWeightOf: association value.
	predecessor := self insertLinkAtEnd: link.
	keyIndex at: association key put: predecessor.
	^ association value! !

!NeoLRUCache methodsFor: 'accessing - statistics' stamp: 'SvenVanCaekenberghe 11/29/2013 21:18'!
hitRatio
	"Return the ratio of hits against total calls I received.
	This will be a number between 0 and 1.
	When I am empty, return 0."
	
	| total |
	total := hits + misses.
	total = 0 ifTrue: [ ^ 0 ].
	^ hits / total! !

!NeoLRUCache methodsFor: 'accessing - statistics' stamp: 'SvenVanCaekenberghe 11/29/2013 21:16'!
hits
	"Return how many hits, requests for keys present I received."

	^ hits! !

!NeoLRUCache methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/29/2013 21:03'!
includesKey: key
	"Return true when the receiver has a value cached for key."
	
	^ keyIndex includesKey: key! !

!NeoLRUCache methodsFor: 'initialize' stamp: 'SvenVanCaekenberghe 11/26/2013 21:26'!
initialize
	super initialize.
	keyIndex := Dictionary new.
	totalWeight := 0.
	maximumWeight := 16.
	computeWeight := [ 1 ].
	hits := misses := 0! !

!NeoLRUCache methodsFor: 'private - linked list' stamp: 'SvenVanCaekenberghe 11/30/2013 19:55'!
insertLinkAtEnd: link
	| predecessor |
	lruTail 
		ifNil: [ 
			lruHead := lruTail := link.
			predecessor := nil ]
		ifNotNil: [
		 	predecessor := lruTail.
			lruTail nextLink: link.
			lruTail := link ].
	link nextLink: nil.
	^ predecessor! !

!NeoLRUCache methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/29/2013 21:22'!
isEmpty
	"Return true when the receiver contains no entries."
	
	^ keyIndex isEmpty! !

!NeoLRUCache methodsFor: 'enumerating' stamp: 'SvenVanCaekenberghe 11/29/2013 21:19'!
keysAndValuesDo: block
	"Execute block with each key and value present in me.
	This will be from least to most recently used."
	
	| link |
	link := lruHead.
	[ link notNil ]
		whileTrue: [ 
			block 
				value: link value key 
				value: link value value.
			link := link nextLink ]! !

!NeoLRUCache methodsFor: 'initialize' stamp: 'SvenVanCaekenberghe 11/29/2013 21:14'!
maximumWeight: limit
	"Set my maximum allowed total weight of all cached values to limit.
	If the total weight is no longer below limit,
	the least recently used key/value pair will be removed.
	The default maximum weight limit is 16."
	
	maximumWeight := limit! !

!NeoLRUCache methodsFor: 'accessing - statistics' stamp: 'SvenVanCaekenberghe 11/29/2013 21:16'!
misses
	"Return how many misses, requests for keys not present I received."
	
	^ misses! !

!NeoLRUCache methodsFor: 'printing' stamp: 'SvenVanCaekenberghe 11/28/2013 09:30'!
printOn: stream
	super printOn: stream.
	stream nextPut: $(.
	stream nextPut: $#; print: self size.
	stream space; print: totalWeight; nextPut: $/; print: maximumWeight.
	stream space; print: computeWeight.
	stream space; print: (self hitRatio * 100.0) rounded ; nextPut: $%.
	stream nextPut: $)! !

!NeoLRUCache methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/1/2013 00:13'!
promoteSuccessorOf: predecessor
	| link  |
	lruHead = lruTail ifTrue: [ ^ nil ].
	predecessor
		ifNil: [ 
			link := lruHead.
			lruHead := link nextLink ]
		ifNotNil: [ 
			link := predecessor nextLink.
			link = lruTail ifTrue: [ ^ predecessor ].
			predecessor nextLink: link nextLink ].
	keyIndex at: link nextLink value key put: predecessor.
	^ self insertLinkAtEnd: link! !

!NeoLRUCache methodsFor: 'removing' stamp: 'SvenVanCaekenberghe 11/29/2013 21:20'!
removeAll
	"Remove all key/value pairs that I currently hold,
	effectiley resetting me, but not my statistics."
	
	keyIndex removeAll.
	totalWeight := 0.
	lruHead := lruTail := nil! !

!NeoLRUCache methodsFor: 'private - linked list' stamp: 'SvenVanCaekenberghe 11/30/2013 19:26'!
removeFirstLink
	^ self removeLink: nil! !

!NeoLRUCache methodsFor: 'removing' stamp: 'SvenVanCaekenberghe 11/29/2013 21:23'!
removeKey: key
	"If I currently cache key, remove the entry.
	Signal a KeyNotFound when I currently do not cache key.
	Return the removed value."
	
	^ self 
		removeKey: key 
		ifAbsent: [ KeyNotFound signalFor: key in: self ]! !

!NeoLRUCache methodsFor: 'removing' stamp: 'SvenVanCaekenberghe 11/30/2013 19:33'!
removeKey: key ifAbsent: block
	"If I currently cache key, remove the entry.
	Execute block when key is currently absent.
	Return the removed value."
	
	^ (self includesKey: key)
		ifTrue: [ 
			| predecessor link |
			predecessor := keyIndex at: key.
			link := self removeLink: predecessor.
			(predecessor notNil and: [ predecessor nextLink notNil ])
				ifTrue: [ keyIndex at: predecessor nextLink value key put: predecessor ].
			self removeWeightOf: link value value.
			keyIndex removeKey: key.
			link value value ]
		ifFalse: block! !

!NeoLRUCache methodsFor: 'private - linked list' stamp: 'SvenVanCaekenberghe 11/30/2013 19:31'!
removeLastLink
	| link |
	link := lruHead.
	lruHead := lruTail := nil.
	^ link! !

!NeoLRUCache methodsFor: 'private - linked list' stamp: 'SvenVanCaekenberghe 11/30/2013 19:31'!
removeLink: predecessor
	| link |
	lruHead = lruTail
		ifTrue: [ 
			^ self removeLastLink ].
	predecessor 
		ifNil: [ 
			link := lruHead.
			lruHead := link nextLink ]
		ifNotNil: [
			link := predecessor nextLink.
			predecessor nextLink: link nextLink ].
	link nextLink: nil.
	^ link! !

!NeoLRUCache methodsFor: 'private' stamp: 'SvenVanCaekenberghe 11/30/2013 19:27'!
removeWeightOf: value 
	| weight |
	weight := computeWeight cull: value.
	totalWeight := totalWeight - weight! !

!NeoLRUCache methodsFor: 'accessing - statistics' stamp: 'SvenVanCaekenberghe 11/29/2013 21:15'!
size
	"Return the count of items currently present."
	
	^ keyIndex size! !

!NeoLRUCache methodsFor: 'accessing - statistics' stamp: 'SvenVanCaekenberghe 11/29/2013 21:15'!
totalWeight
	"Return the total weight of all cached values currently present."
	
	^ totalWeight! !

!NeoLRUCache methodsFor: 'private' stamp: 'SvenVanCaekenberghe 11/28/2013 23:01'!
validateInvariantWith: assertable
	keyIndex keysAndValuesDo: [ :key :predecessor |
		predecessor 
			ifNil: [ 
				assertable assert: lruHead value key = key ]
			ifNotNil: [ 
				assertable assert: predecessor nextLink value key = key ] ].
	self keysAndValuesDo: [ :key :value |
		(keyIndex at: key)
			ifNil: [ 
				assertable assert: lruHead value value = value.
				assertable assert: lruHead value key = key ]
			ifNotNil: [ :predecessor | 
				assertable assert: predecessor nextLink value value = value.
				assertable assert: predecessor nextLink value key = key ] ]! !
